import os
import requests
import json

# Extract Kentik API settings from environment variables
API_URL  = os.environ["KENTIK_API_URL"]
API_EMAIL= os.environ["KENTIK_API_EMAIL"]
API_TOKEN= os.environ["KENTIK_API_TOKEN"]

# Kentik report period (in minutes)
KENTIK_REPORTING_PERIOD=60
# Kentik flow aggregation window (in minutes -- lower values improve accuracy at the expense of processing time)
KENTIK_FLOW_AGGR_WINDOW=5

# aggregate ingress (source) and egress (destination) on the Kentik 'Site'-level
INGRESS_DIMENSION="i_device_site_name"
EGRESS_DIMENSION ="i_ult_exit_site"

###############################################################################
# Fetch traffic matrix from the Kentik API
###############################################################################

def kentik_api_query(payload):
    # Headers for authentication
    headers = {
        "Content-Type": "application/json",
        "X-CH-Auth-Email": API_EMAIL,
        "X-CH-Auth-API-Token": API_TOKEN
    }
    response = requests.post(API_URL, headers=headers, json=payload)

    if response.status_code == 200:
        return response.json()
    else:
        print(f"Error fetching data from Kentik API: {response.status_code} - {response.text}")
        return None

# Example in Kentik dashboard: https://portal.kentik.eu/v4/core/explorer/5787b62a7e8ae1ef7d399e08cdea2800
def fetch_kentik_trafic_matrix():
    # JSON query payload
    payload = {
      "version": 4,
      "queries": [
        {
          "bucket": "Table",
          "isOverlay": False,
          "query": {
            "all_devices": True,
            "aggregateTypes": [
              "max_in_bits_per_sec"
            ],
            "depth": 350,
            "topx": 350,                        # 350=max supported by Kentik
            "device_name": [],
            "fastData": "Auto",
            "lookback_seconds": 60 * KENTIK_REPORTING_PERIOD,
#            "starting_time": null,
#            "ending_time": null,
            "matrixBy": [],
            "metric": [
              "in_bytes"
            ],
            "minsPolling": KENTIK_FLOW_AGGR_WINDOW,
            "forceMinsPolling": True,
            "outsort": "max_in_bits_per_sec",
            "viz_type": "table",
            "aggregates": [
              {
                "value": "max_in_bits_per_sec",
                "column": "f_sum_in_bytes",
                "fn": "max",
                "label": "Bits/s Sampled at Ingress Max",
                "unit": "in_bytes",
                "parentUnit": "bytes",
                "group": "Bits/s Sampled at Ingress",
                "origLabel": "Max",
                "sample_rate": 1,
                "raw": True,
                "name": "max_in_bits_per_sec"
              }
            ],
            "filters": {
              "connector": "All",
              "filterGroups": [
                {
                  "connector": "All",
                  "filters": [
                    {
                      "filterField": EGRESS_DIMENSION,
                      "metric": "",
                      "aggregate": "",
                      "operator": "<>",
                      "filterValue": "",            # Filter unassigned/unknown destinations
                      "rightFilterField": ""
                    }
                  ],
                  "filterGroups": []
                }
              ]
            },
            "dimension": [
              INGRESS_DIMENSION,
              EGRESS_DIMENSION
            ]
          }
        }
      ]
    }

    response = kentik_api_query(payload)
    return response

###############################################################################
# Transform Kentik data to traffic matrices (one matrix per timestamp)
###############################################################################

def kentik_to_traffic_matrices(json_data, nodes):
    """
    Convert the given JSON structure returned by Kentik into a dictionary
    keyed by timestamp. For each timestamp, we store a nested dict of:
        traffic_matrices[timestamp][ingress][egress] = traffic_rate_Mbps
    """

    # We'll gather all flows in the JSON
    data_entries = json_data["results"][0]["data"]

    # This will be our main lookup: { ts -> {ingress -> { egress -> rate}} }
    traffic_matrices = {}

    for entry in data_entries:
        ingress= entry[INGRESS_DIMENSION]
        egress = entry[EGRESS_DIMENSION]

        # skip unknown ingress and/or egress nodes
        if ingress not in nodes:
            continue
        if egress not in nodes:
            continue

        # timeSeries -> in_bits_per_sec -> flow => list of [timestamp, in_bits_per_sec, ignored_interval]
        flow_list = entry["timeSeries"]["in_bits_per_sec"]["flow"]
        for (timestamp, bits_per_sec, _ignored_interval) in flow_list:
            traffic_matrices.setdefault(timestamp, {})
            traffic_matrices[timestamp].setdefault(ingress, {})
            traffic_matrices[timestamp][ingress][egress] = int(bits_per_sec / 1_000_000)     # convert bps to Mbps

    return traffic_matrices