import math
import itertools
import networkx as nx      # https://networkx.org
import pandas as pd        # https://pandas.pydata.org
import multiprocessing
from datetime import datetime, timezone
from pathlib import Path
import KentikTrafficMatrix as ktm
import NetworkTopology as nettopo

###############################################################################
# 1) INPUT DATA SECTION
###############################################################################

# If normal threshold not specified on a link, default to 0.40 (allowing upto 40% link utilization in normal non-failure scenario)
DEFAULT_NORMAL_THRESHOLD = 0.40
# If failure threshold not specified on a link, default to 0.80 (allowing upto 80% link utilization in worst-case failure scenario)
DEFAULT_FAILURE_THRESHOLD= 0.80

ENABLED_SCENARIOS=["normal","onelinkfail","twolinkfail","onenodefail","nodelinkfail"]

# Scenarios triggering a 'MUST UPGRADE' if threshold is exceeded 
MUST_UPGRADE_SCENARIOS=["normal","onelinkfail","onenodefail"]
# Scenarios triggering a 'SHOULD UPGRADE' if threshold is exceeded 
SHOULD_UPGRADE_SCENARIOS=["twolinkfail","nodelinkfail"]

RAW_REPORT_DIRECTORY="/Users/daniel.verlouw/Desktop/rawreports"
RAW_REPORT_FILE_PREFIX="raw_capacityreport_"
RAW_REPORT_FILE_SUFFIX="csv"

# ISO 8601 basic timestamp format (YYYYMMDDTHHMMZ)
ISO8601_FORMAT = '%Y%m%dT%H%MZ'
ISO8601_REGEXP = r'\d{4}\d{2}\d{2}T\d{2}\d{2}Z'

###############################################################################
# 2) Build a local NetworkX multi-directional graph for scenario computations
###############################################################################

def build_graph(nodes, links):
    # MultiDiGraph supports multiple edges (ECMP links) between two nodes
    # and is directional, allowing modelling of Full-Duplex links and taking
    # traffic direction into account
    graph= nx.MultiDiGraph()
    graph.add_nodes_from(nodes)

    for link in links:
        # ( linkID, nodeA, nodeB, igp_metric, capacity, srlg_list, [norm_thr], [fail_thr] )
        lID  = link[0]
        nodeA= link[1]
        nodeB= link[2]
        igp  = link[3]

        # Add Full-Duplex edges in both directions with linkID
        # assumes the IGP metric is configured symmetrically
        graph.add_edge(nodeA,nodeB,key=lID, igpmetric=igp)
        graph.add_edge(nodeB,nodeA,key=lID, igpmetric=igp)

    return graph

###############################################################################
# 3) Helper functions to ease lookups for core links and SRLGs
###############################################################################

def build_corelink_map(links):
    # Creating a lookup dictionary for fast lookup by linkID
    corelink_dict = {link[0]: link for link in links}
    return corelink_dict

def build_srlg_map(links):
    # Maps SRLG name to one or more core links (linkIDs)
    srlg_map={}

    for link in links:
        # ( linkID, nodeA, nodeB, igp_metric, capacity, srlg_list, [norm_thr], [fail_thr] )
        lID  = link[0]
        nodeA= link[1]
        nodeB= link[2]
        igp  = link[3]

        if len(link)>5 and isinstance(link[5], list):
            for srlg in link[5]:
                srlg_map.setdefault(srlg, set())
                srlg_map[srlg].add(lID)

    return srlg_map

###############################################################################
# 4) Node failover logic - redistribute traffic to other core nodes
###############################################################################

def redistribute_local_traffic(t_xx, ratio_map):
    """
    We treat local traffic T[X,X] as two endpoints at X.
    Each endpoint re-homes to neighbor i or j with probability ratio_map[i], ratio_map[j].
    => fraction = ratio_map[i]*ratio_map[j]*t_xx => T[i,j].
    If i=j, it's T[i,i] (local on neighbor i).
    Returns local_matrix[i][j] = fraction (Mbps).
    """
    local_matrix={}
    nbrs= list(ratio_map.keys())
    for i in nbrs:
        local_matrix.setdefault(i,{})
        for j in nbrs:
            local_matrix[i][j]= ratio_map[i]* ratio_map[j]* t_xx

    return local_matrix

def build_traffic_with_node_fail(baseTraffic, nodefail_ratios, failingNode):
    """
    Return scenario-specific traffic matrix in Mbps,
    re-homing T[failingNode,*], T[*,failingNode], T[failingNode,failingNode].
    """    
    scenario_traffic={}
    ratio_map= nodefail_ratios.get(failingNode,{})

    for s in baseTraffic:
        scenario_traffic[s]={}
        for d in baseTraffic[s]:
            val= baseTraffic[s][d]
            if val<=0:
                scenario_traffic[s][d]=0
                continue

            # traffic is local - T[failingNode,failingNode]
            if s==failingNode and d==failingNode:
                scenario_traffic[s][d]=0
                mat= redistribute_local_traffic(val, ratio_map)
                for i in mat:
                    for j in mat[i]:
                        portion= mat[i][j]
                        if portion>0:
                            scenario_traffic.setdefault(i,{})
                            scenario_traffic[i][j]= scenario_traffic[i].get(j,0)+ portion

            # T[failingNode,*]
            elif s==failingNode:
                scenario_traffic[s][d]=0
                for nbr,r in ratio_map.items():
                    scenario_traffic.setdefault(nbr,{})
                    scenario_traffic[nbr][d]= scenario_traffic[nbr].get(d,0)+ val*r

            # T[*,failingNode]
            elif d==failingNode:
                scenario_traffic[s][d]=0
                for nbr,r in ratio_map.items():
                    scenario_traffic[s][nbr]= scenario_traffic[s].get(nbr,0)+ val*r

            # Unaffected
            else:
                scenario_traffic[s][d]= val

    return scenario_traffic

###############################################################################
# 5) Remove SRLG or corelink from graph
###############################################################################

def remove_corelink(graph, removeLink):
    edges_rm=[]
    for (u,v,lID) in graph.edges(keys=True):
        if lID==removeLink:
            edges_rm.append((u,v,lID))
    for e in edges_rm:
        graph.remove_edge(*e)

def remove_corelink_or_srlg(graph, failedElement, srlg_map):
    etype, data= failedElement
    if etype=="LINK":
        linkID= data
        remove_corelink(graph, linkID)

    elif etype=="SRLG":
        if data in srlg_map:
            for linkID in srlg_map[data]:
                remove_corelink(graph, linkID)

###############################################################################
# 6) Compute scenario load in MultiDiGraph
###############################################################################

def compute_scenario_load(graph, scenario_traffic):
    # initialize empty usage map
    usageMap={}
    for(u,v,lID) in graph.edges(keys=True):
        usageMap[(u,v,lID)] = 0.0

    # walk over [ingress][egress] traffic demands
    for s in scenario_traffic:
        for d in scenario_traffic[s]:
            demand_mbps= scenario_traffic[s][d]
            if demand_mbps<=0 or s==d or (s not in graph) or (d not in graph):
                continue
            demand_gbps= demand_mbps/1000.0    # convert Mbps to Gbps

            # Run Dijkstra algorithm
            try:
                best_dist= nx.dijkstra_path_length(graph, s, d, weight='igpmetric')
            except nx.NetworkXNoPath:
                continue

            # Enumerate all shortest paths from (s)ource to (d)estination
            all_paths= list(nx.all_shortest_paths(graph, s, d, weight='igpmetric'))
            valid=[]
            for path in all_paths:
                dist_sum=0
                for i in range(len(path)-1):
                    p= path[i]
                    q= path[i+1]
                    # pick minimal igp among parallel edges p->q
                    best_igp= math.inf
                    for k2 in graph[p][q]:
                        igp= graph[p][q][k2]['igpmetric']
                        if igp<best_igp:
                            best_igp= igp
                    dist_sum+= best_igp

                if math.isclose(dist_sum,best_dist):
                    valid.append(path)

            if not valid:
                continue

            # Now equally divide traffic demand across all shortest paths (ECMP)
            share= demand_gbps/ len(valid)
            for path in valid:
                for i in range(len(path)-1):
                    p= path[i]
                    q= path[i+1]
                    # pick minimal link p->q
                    best_igp= math.inf
                    best_k= None
                    for k2 in graph[p][q]:
                        igp= graph[p][q][k2]['igpmetric']
                        if igp<best_igp:
                            best_igp= igp
                            best_k= k2
                    usageMap[(p,q,best_k)]+= share

    return usageMap

###############################################################################
# 7) Individual Normal/Failure Scenario Runners
###############################################################################

def run_normal_scenario(graph, trafficMatrix):
    """
    No failures => just compute load with the base traffic.
    """        
    return compute_scenario_load(graph, trafficMatrix)

def run_onenode_failure(graph, trafficMatrix, nodefail_ratios, nodeFail):
    """
    Remove nodeFail from graph, shift its traffic T[nodeFail,*], T[*,nodeFail], T[nodeFail,nodeFail].
    """        
    if nodeFail in graph:
        graph.remove_node(nodeFail)
    scenario_traffic= build_traffic_with_node_fail(trafficMatrix, nodefail_ratios, nodeFail)
    return compute_scenario_load(graph, scenario_traffic)

def run_onelink_failure(graph, trafficMatrix, srlg_map, linkFail):
    """
    linkFail = ("LINK",(u,v)) or ("SRLG",{(x,y),...}).
    Remove link from a copy of graph, then compute load with base traffic.
    """        
    remove_corelink_or_srlg(graph, linkFail, srlg_map)
    return compute_scenario_load(graph, trafficMatrix)

def run_twolink_failure(graph, trafficMatrix, srlg_map, linkFail1, linkFail2):
    """
    linkFail1, linkFail2 each = ("LINK",(u,v)) or ("SRLG",{(x,y),...}).
    Remove them from a copy of graph, then compute load with base traffic.
    """        
    remove_corelink_or_srlg(graph, linkFail1, srlg_map)
    remove_corelink_or_srlg(graph, linkFail2, srlg_map)
    return compute_scenario_load(graph, trafficMatrix)

def run_nodelink_failure(graph, trafficMatrix, srlg_map, nodefail_ratios, nodeFail, linkFail):
    # first remove node and shift traffic
    if nodeFail in graph:
        graph.remove_node(nodeFail)
    scenario_traffic= build_traffic_with_node_fail(trafficMatrix, nodefail_ratios, nodeFail)
    # then remove link
    remove_corelink_or_srlg(graph, linkFail, srlg_map)
    return compute_scenario_load(graph, scenario_traffic)

###############################################################################
# 8) Single scenario task runner
###############################################################################

def expand_link_name(corelink_map, linkFail):
    """Helper function--return a string describing one scenario element (LINK linkID or SRLG name)."""
    etype,data= linkFail
    if etype=="LINK":
        # get corelink details by ID
        td= corelink_map.get(data)
        nodeA= td[1]
        nodeB= td[2]
        return f"LINK_{data}/{nodeA}-{nodeB}"

    elif etype=="SRLG":
        return f"SRLG_{data}"

    else:
        return str(linkFail)

def compute_scenario_task(args):
    """
    Each scenario runs as a single task
    """
    (graph_orig, corelink_map, srlg_map, nodefail_ratios, ts, trafficMatrix, scenarioType, scenarioData)= args

    # 1) first, create a copy of graph so in-memory copy of original is not modified
    graph=graph_orig.copy()

    # 2) run selected normal or failure scenario
    usageMap={}
    scenarioName=""
    if scenarioType=="normal":
        usageMap= run_normal_scenario(graph, trafficMatrix)
        scenarioName="normal"

    elif scenarioType=="onenodefail":
        nodeFail= scenarioData
        usageMap= run_onenode_failure(graph, trafficMatrix, nodefail_ratios, nodeFail)
        scenarioName= f"NODE_{nodeFail}"

    elif scenarioType=="onelinkfail":
        linkFail= scenarioData
        usageMap= run_onelink_failure(graph, trafficMatrix, srlg_map, linkFail)
        scenarioName= f"{expand_link_name(corelink_map, linkFail)}"

    elif scenarioType=="twolinkfail":
        (linkFail1,linkFail2)= scenarioData
        usageMap= run_twolink_failure(graph, trafficMatrix, srlg_map, linkFail1, linkFail2)
        scenarioName= f"{expand_link_name(corelink_map, linkFail1)} + {expand_link_name(corelink_map, linkFail2)}"

    elif scenarioType=="nodelinkfail":
        (nodeFail,linkFail)= scenarioData
        usageMap= run_nodelink_failure(graph, trafficMatrix, srlg_map, nodefail_ratios, nodeFail, linkFail)
        scenarioName= f"NODE_{nodeFail} + {expand_link_name(corelink_map, linkFail)}"

    else:
        usageMap={}
        scenarioName="unknown"

    return (ts, scenarioType, scenarioName, usageMap)


###############################################################################
# 9) Build list of tasks to run in parallel
###############################################################################

def build_parallel_tasks(graph, nodes, nodefail_ratios, corelink_map, srlg_map, traffic_matrices):
    # enumerate link failure scenarios => link linkIDs + srlg
    linkFailElements=[]
    for link_id, _link_data in corelink_map.items():
        linkFailElements.append(("LINK", link_id))

    for srlg_name, _srlg_data in srlg_map.items():
        linkFailElements.append(("SRLG", srlg_name))

    tasks=[]
    for ts in traffic_matrices.keys():
        trafficMatrix= traffic_matrices[ts]

        # add normal non-failure scenario
        if "normal" in ENABLED_SCENARIOS:
            tasks.append((graph, corelink_map, srlg_map, nodefail_ratios, ts, trafficMatrix, "normal", None))

        # add single link fail scenarios
        if "onelinkfail" in ENABLED_SCENARIOS:
            for linkFail in linkFailElements:
                tasks.append((graph, corelink_map, srlg_map, nodefail_ratios, ts, trafficMatrix, "onelinkfail", linkFail))

        # add unique two link fail scenarios
        if "twolinkfail" in ENABLED_SCENARIOS:
            fe_twofail_combos= list(itertools.combinations(linkFailElements,2))
            for linkFail1,linkFail2 in fe_twofail_combos:
                tasks.append((graph, corelink_map, srlg_map, nodefail_ratios, ts, trafficMatrix, "twolinkfail", (linkFail1,linkFail2)))

        for nodeFail in nodes:
            # add node fail scenarios
            if "onenodefail" in ENABLED_SCENARIOS:
                tasks.append((graph, corelink_map, srlg_map, nodefail_ratios, ts, trafficMatrix, "onenodefail", nodeFail))

            # add node + link fail scenarios
            if "nodelinkfail" in ENABLED_SCENARIOS:
                for linkFail in linkFailElements:
                    tasks.append((graph, corelink_map, srlg_map, nodefail_ratios, ts, trafficMatrix, "nodelinkfail", (nodeFail,linkFail)))

    return tasks

###############################################################################
# 10) Report generator
###############################################################################

def link_usage_aggregator(corelink_map, results):
    # initialize empty link usage aggregator
    aggLinkUsage={}
    for link_id, _link_data in corelink_map.items():
        aggLinkUsage[link_id]= {}
        for scenarioType in ENABLED_SCENARIOS:
            aggLinkUsage[link_id][scenarioType] = (0.0,pd.NaT,None)

    # unify results across all tasks - for each scenario, record worst-case (highest) load on each link
    for (ts, scenarioType, scenarioName, usageMap) in results:
        for (u,v,lID) in usageMap:
            usage= usageMap[(u,v,lID)]
            oldUsage, oldTs, oldSc= aggLinkUsage[lID][scenarioType]
            if usage> oldUsage:
                aggLinkUsage[lID][scenarioType]= (usage, ts, f"{scenarioName}")

    # return aggLinkUsage[linkID][scenarioType] => (usage,ts,scenarioName)
    return aggLinkUsage

def build_raw_report(corelink_map, aggLinkUsage):
    final_report_raw=[]

    # Helper print functions
    def exceed(scenario, thr, capacity):
        return aggLinkUsage[lID][scenario][0]>thr*capacity if scenario in ENABLED_SCENARIOS else pd.NA

    def usagePct(scenario, capacity):
        # (usage,ts,scenarioName)
        return aggLinkUsage[lID][scenario][0]/capacity if scenario in ENABLED_SCENARIOS else pd.NA

    def usageGbps(scenario):
        return aggLinkUsage[lID][scenario][0] if scenario in ENABLED_SCENARIOS else pd.NA

    def scenarioTimestamp(scenario):
        return pd.to_datetime(aggLinkUsage[lID][scenario][1], unit='ms', errors='coerce', utc=True) if scenario in ENABLED_SCENARIOS else pd.NaT

    def failScenarioName(scenario):
        return aggLinkUsage[lID][scenario][2] if scenario in ENABLED_SCENARIOS else None

    def scenarioEnabled(scenario):
        return scenario in ENABLED_SCENARIOS

    # Iterate over each core link
    for link_id, link_data in corelink_map.items():
        # ( linkID, nodeA, nodeB, igp_metric, capacity, srlg_list, [norm_thr], [fail_thr] )
        lID     = link_data[0]
        nodeA   = link_data[1]
        nodeB   = link_data[2]
        capacity= link_data[4]
        # parse thresholds if present, else use defaults
        norm_thr= link_data[6] if len(link_data)>6 and isinstance(link_data[6], (float,int)) else DEFAULT_NORMAL_THRESHOLD
        fail_thr= link_data[7] if len(link_data)>7 and isinstance(link_data[7], (float,int)) else DEFAULT_FAILURE_THRESHOLD

        # threshold check
        mustUpgrade=False
        shouldUpgrade=False
        for scenarioType in ENABLED_SCENARIOS:
            (usage,ts,scenarioName) = aggLinkUsage[lID][scenarioType]
            thr= 1
            if scenarioType=="normal" and norm_thr>0:
                thr=norm_thr
            elif "fail" in scenarioType and fail_thr>0:
                thr=fail_thr

            if usage> thr*capacity and scenarioType in MUST_UPGRADE_SCENARIOS:
                mustUpgrade=True
            elif usage> thr*capacity and scenarioType in SHOULD_UPGRADE_SCENARIOS:
                shouldUpgrade=True

        # print one row per core link
        final_report_raw.append({
            "ID":                           lID,
            "NodeA":                        nodeA,
            "NodeB":                        nodeB,
            "CapacityGbps":                 capacity,

            "ConfNormalThrPct":             norm_thr,
            "ConfNormalThrGbps":            norm_thr*capacity,

            "ConfFailThrPct":               fail_thr,
            "ConfFailThrGbps":              fail_thr*capacity,

            "NormalEnabled":                scenarioEnabled("normal"),
            "NormalUsagePct":               usagePct("normal", capacity),
            "NormalUsageGbps":              usageGbps("normal"),
            "NormalUsageExceed":            exceed("normal", norm_thr, capacity),
            "NormalDateTime":               scenarioTimestamp("normal"),

            "1LinkFailEnabled":             scenarioEnabled("onelinkfail"),
            "1LinkFailScenario":            failScenarioName("onelinkfail"),
            "1LinkFailUsagePct":            usagePct("onelinkfail", capacity),
            "1LinkFailUsageGbps":           usageGbps("onelinkfail"),
            "1LinkFailUsageExceed":         exceed("onelinkfail", fail_thr, capacity),
            "1LinkFailUsageTime":           scenarioTimestamp("onelinkfail"),

            "2LinkFailEnabled":             scenarioEnabled("twolinkfail"),
            "2LinkFailScenario":            failScenarioName("twolinkfail"),
            "2LinkFailUsagePct":            usagePct("twolinkfail", capacity),
            "2LinkFailUsageGbps":           usageGbps("twolinkfail"),
            "2LinkFailUsageExceed":         exceed("twolinkfail", fail_thr, capacity),
            "2LinkFailUsageTime":           scenarioTimestamp("twolinkfail"),

            "1NodeFailEnabled":             scenarioEnabled("onenodefail"),
            "1NodeFailScenario":            failScenarioName("onenodefail"),
            "1NodeFailUsagePct":            usagePct("onenodefail", capacity),
            "1NodeFailUsageGbps":           usageGbps("onenodefail"),
            "1NodeFailUsageExceed":         exceed("onenodefail", fail_thr, capacity),
            "1NodeFailUsageTime":           scenarioTimestamp("onenodefail"),

            "Node+1LinkFailEnabled":        scenarioEnabled("nodelinkfail"),
            "Node+1LinkFailScenario":       failScenarioName("nodelinkfail"),
            "Node+1LinkFailUsagePct":       usagePct("nodelinkfail", capacity),
            "Node+1LinkFailUsageGbps":      usageGbps("nodelinkfail"),
            "Node+1LinkFailUsageExceed":    exceed("nodelinkfail", fail_thr, capacity),
            "Node+1LinkFailUsageTime":      scenarioTimestamp("nodelinkfail"),

            "MustUpgrade":                  mustUpgrade,
            "ShouldUpgrade":                shouldUpgrade
        })

    df_raw= pd.DataFrame(final_report_raw, columns=["ID","NodeA","NodeB","CapacityGbps",
                                            "ConfNormalThrPct","ConfNormalThrGbps",
                                            "ConfFailThrPct","ConfFailThrGbps",
                                            "NormalEnabled","NormalUsagePct","NormalUsageGbps","NormalUsageExceed","NormalDateTime",
                                            "1LinkFailEnabled","1LinkFailScenario","1LinkFailUsagePct","1LinkFailUsageGbps","1LinkFailUsageExceed","1LinkFailUsageTime",
                                            "2LinkFailEnabled","2LinkFailScenario","2LinkFailUsagePct","2LinkFailUsageGbps","2LinkFailUsageExceed","2LinkFailUsageTime",
                                            "1NodeFailEnabled","1NodeFailScenario","1NodeFailUsagePct","1NodeFailUsageGbps","1NodeFailUsageExceed","1NodeFailUsageTime",
                                            "Node+1LinkFailEnabled","Node+1LinkFailScenario","Node+1LinkFailUsagePct","Node+1LinkFailUsageGbps","Node+1LinkFailUsageExceed","Node+1LinkFailUsageTime",
                                            "MustUpgrade","ShouldUpgrade",])
    df_raw.set_index("ID", inplace=True)

    return df_raw

def store_raw_report(df_raw):
    directory = Path(RAW_REPORT_DIRECTORY)
    directory.mkdir(parents=True, exist_ok=True)  # Create directory if it doesn't exist

    # Create a ISO8601 basic format UTC timestamped filename
    timestamp = datetime.now(timezone.utc).strftime(ISO8601_FORMAT)
    filename = f'{RAW_REPORT_FILE_PREFIX}{timestamp}.{RAW_REPORT_FILE_SUFFIX}'
    filepath = directory / filename

    df_raw.to_csv(filepath, sep=',', encoding='utf-8', date_format=ISO8601_FORMAT, header=True)


###############################################################################
# MAIN
###############################################################################

def main():
    # Record and display the start time of the script
    start_time = datetime.now(timezone.utc)
    print(f"===============================================================")
    print("Execution started at:", start_time.strftime(ISO8601_FORMAT))

    # 1) query Kentik to build traffic demand matrices
    print(f"Querying Kentik API for traffic demand matrix...")
    kentik_json = ktm.fetch_kentik_trafic_matrix()
    print(f"Processing traffic demand matrices from Kentik data...")
    traffic_matrices= ktm.kentik_to_traffic_matrices(kentik_json, nettopo.NODES)       # returns traffic_matrices[timestamp][ingress][egress] = traffic_rate_Mbps

    # 2) build graph and lookup data structures
    graph= build_graph(nettopo.NODES, nettopo.CORELINKS)
    corelink_map= build_corelink_map(nettopo.CORELINKS)
    srlg_map= build_srlg_map(nettopo.CORELINKS)

    # 3) define tasks to run in parallel
    tasks= build_parallel_tasks(graph, nettopo.NODES, nettopo.NODE_FAILOVER_RATIOS, corelink_map, srlg_map, traffic_matrices)
    print(f"Number of traffic simulation tasks: {len(tasks)} across {len(traffic_matrices)} timeslots with {len(tasks)/len(traffic_matrices):.0f} scenarios each")

    # 4) now run all tasks in a single global pool
    print(f"Executing parallel traffic simulation tasks...")
    with multiprocessing.Pool() as pool:
        tasks_results= pool.map(compute_scenario_task, tasks)

    # 5) aggregate link usage across all tasks, recording the worst-case (highest) load on each link
    aggLinkUsage= link_usage_aggregator(corelink_map, tasks_results)
 
    # 6) create and store final report
    raw_report= build_raw_report(corelink_map, aggLinkUsage)
    store_raw_report(raw_report)

    # Display the end time, and total execution duration
    end_time = datetime.now(timezone.utc)
    duration = end_time - start_time
    print("Execution ended at:", end_time.strftime(ISO8601_FORMAT))
    print("Total duration in seconds:", duration.total_seconds())
    print(f"===============================================================")

if __name__=="__main__":
    main()
